/*****************************************************************************
 *  FILE   : dcali.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: dcali.c,v 1.2 1997/07/14 17:43:16 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * SWITCHES
 *****************************************************************************/
/* #define PRIMITIVE_CYCLE *//* for quality testing only: not optimized */

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* fprintf(),fputc() */
#include <float.h> /* DBL_MAX */
#include <sys/types.h> /* clock_t */
#include <sys/times.h> /* struct tms,times() */

#include "global.h" /* data types,ALLOC(),FREE(),MIN(),MAX(),CLK_TCK */
#include "dypr.h" /* calc_column(),calc_matrix() */
#include "seqs.h" /* function prototypes */
#include "dcali.h" /* function prototypes */

/*****************************************************************************
 * MACROS
 *****************************************************************************/
#define SWAP_SEQS(seq,s1,s2) swap_s=seq.s[s1]; seq.s[s1]=seq.s[s2];\
                                                seq.s[s2]=swap_s;\
                             swap_len=seq.len[s1]; seq.len[s1]=seq.len[s2];\
                                                    seq.len[s2]=swap_len;\
                             for(swap_p=0; swap_p<seq.num; ++swap_p) {\
                               swap_wf = seq.wf[s1][swap_p];\
                               seq.wf[s1][swap_p] = seq.wf[s2][swap_p];\
                               seq.wf[s2][swap_p] = swap_wf;\
                             }\
                             for(swap_p=0; swap_p<seq.num; ++swap_p) {\
                               swap_wf = seq.wf[swap_p][s1];\
                               seq.wf[swap_p][s1] = seq.wf[swap_p][s2];\
                               seq.wf[swap_p][s2] = swap_wf;\
                             }
#define SWAP_ALI(al,a1,a2) swap_a=al.a[a1]; al.a[a1]=al.a[a2]; al.a[a2]=swap_a
#define SWAP_TUPLE(t,e1,e2) swap_e=t[e1]; t[e1]=t[e2]; t[e2]=swap_e

/*****************************************************************************
 * GLOBAL VARIABLES
 *****************************************************************************/
sequence swap_s,swap_a;
int swap_len,swap_p,swap_e;
float swap_wf;

int i, /* calc_Cmin() */
    **score; /* calc_Chat() */
column **col; /* calc_Chat(),calc_Cmin() */
matrix **m; /* calc_Chat(),calc_Cmin() */
tuple lo,uo,olo,olen, /* calc_Chat(),calc_Cmin() */
      tmp_cut,lower_tmp_cut,upper_tmp_cut; /* calc_Cmin() */
double Cmin, Chat, /* calc_Chat(),calc_Cmin() */
       *c,act_cost; /* calc_Cmin() */
BOOL *down; /* calc_Cmin() */

#ifdef WRITE_STATISTICS
clock_t timeChat,timeCmin;
#endif

/*****************************************************************************
 * EXTERNAL FUNCTIONS
 *****************************************************************************/
/*===========================================================================*/
/* -- CALC_CHAT --
 * Calculate an estimate 'Chat' with corresponding cut positions 'cut' */
/*---------------------------------------------------------------------------*/
#include "calc_Chat.c"

/*===========================================================================*/
/* -- CALC_CMIN --
 * Calculate minimal additional cost 'Cmin' with corresponding cut positions
 * 'cut' */
/*---------------------------------------------------------------------------*/
#include "calc_Cmin.c"

/*****************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************/

/*===========================================================================*/
/* -- CALC_CUT --
 * Calculate cut positions 'cut' between lower and upper bounds ('l' and 'u'
 * resp.) of global sequences 'seqs'. 's' is the number of the sequence to be
 * cut at its midpoint */
/*---------------------------------------------------------------------------*/
void calc_cut(tuple l, tuple u, tuple cut, int s)
{
  /* Search within: l <= cut <= u.
   * Initially: 0 <= cut <= u-l.
   * After calculating offsets: 0 <= lo <= cut <= uo <= u-l
   *                       and: lo <= tmp_cut <= uo.
   * Later add l to the cut points.
   */

  extern sequences seqs;
  extern parameters par;
  int p,q, j;
#ifdef WRITE_STATISTICS
  struct tms start,stop;
#endif

  /* swap sequences */
  if(!par.spantree) {
    SWAP_SEQS(seqs,0,s);
    SWAP_TUPLE(l,0,s);
    SWAP_TUPLE(u,0,s);
    /* SWAP_TUPLE(cut,0,s); not necessary, since empty, yet */
  }

  /* calculate estimate Chat */
#ifdef WRITE_STATISTICS
  times(&start);
#endif
  calc_Chat(l,u,cut);
#ifdef WRITE_STATISTICS
  times(&stop);
  timeChat += stop.tms_utime-start.tms_utime;
#endif

  if(par.o) {
    fputc('*',stderr);
    fflush(stderr);
  }

  if(!par.a) {
    /* calculate minimal additional cost Cmin */
#ifdef WRITE_STATISTICS
    times(&start);
#endif
    calc_Cmin(l,u,cut);
#ifdef WRITE_STATISTICS
    times(&stop);
    timeCmin += stop.tms_utime-start.tms_utime;
#endif
  } /* if !par.a */

  if(par.o) {
    fputc('*',stderr);
    fflush(stderr);
  }

  /* free matrices (were allocated in calc_Chat()) */
  for(p=2; p<seqs.num; ++p)
    for(q=1; q<p; ++q) {
      for(j=0; j<olen[p]; ++j) /* original lengths */
        FREE(m[p][q][j]);
      FREE(m[p][q]);
    }

  /* increment cut by left bound */
  for(p=0; p<seqs.num; ++p)
    cut[p] += l[p];

  /* swap sequences back */
  if(!par.spantree) {
    SWAP_SEQS(seqs,0,s);
    SWAP_TUPLE(l,0,s);
    SWAP_TUPLE(u,0,s);
    SWAP_TUPLE(cut,0,s);
  }
} /* calc_cut() */

/*===========================================================================*/
/* -- DCA --
 * Recursive divide and conquer alignment function */
/*---------------------------------------------------------------------------*/
void dca(tuple l, tuple u, int level)
{
  extern parameters par;
  extern sequences seqs;
  extern alignment ali;

  tuple c;
  int p,pmax, len,maxlen, cut_point;
  sequences tmp_s;
  alignment tmp_a;

  for(pmax=0,maxlen=u[0]-l[0],p=1; p<seqs.num; ++p) /* find longest sequence */
    if((len=u[p]-l[p]) > maxlen) {
      pmax = p;
      maxlen = len;
    }

  if(maxlen>par.l && (!par.spantree || u[0]-l[0]>1)) { /* check L */
    ALLOC(c,int,seqs.num);
    calc_cut(l,u,c,pmax); /* cut sequences */
#ifndef FIRST_CUT_ONLY
    dca(l,c,level+1);
    cut_point = ali.len;
    dca(c,u,level+1);
    ali.a[seqs.num][cut_point] = level; /* set level entry */
#endif
    FREE(c);
  }
#ifndef DO_NOT_CALCULATE_RESTCUBES
  else { /* align remaining subsequences optimally (call msa) */
    copy_seqs(&seqs,l,u,&tmp_s);
    msa(&tmp_s,&tmp_a,par.free_shift && l[pmax]==0,
                      par.free_shift && u[pmax]==seqs.len[pmax]);
    append_ali(&tmp_a,0,tmp_a.len,&ali);
    free_ali(&tmp_a); /* was allocated in msa() */
    free_seqs(&tmp_s);
  }
#endif
} /* dca() */

/*===========================================================================*/
/* -- REALIGN --
 * Do "windowing" with window size 'windowsize'  over cut positions (i.e.
 * sites with non-negative 'level' entry) of global alignment 'ali' */
/*---------------------------------------------------------------------------*/
void realign(int windowsize)
{
  extern parameters par;
  extern sequences seqs;
  extern alignment ali;

  int i,j,k,l,r,left,right;
  sequences tmp_s;
  alignment copy_a,new_a;

  l = windowsize/2;
  r = windowsize-l;

  if(par.o) {
    fprintf(stderr,"Re-aligning Windows.\n");
    for(i=0; i<ali.len; ++i)
      if(ali.a[ali.num][i] >= 0)
        fprintf(stderr,".");
    fprintf(stderr,"\n");
    fflush(stderr);
  }

  /* allocate memory for copies of window-subsequences */
  alloc_seqs(&seqs,&tmp_s);
  copy_ali(&ali,&copy_a); /* make copy of alignment */

  ali.len = 0; /* reset alignment, msa_approx unchanged */
  for(ali.len=j=i=0; i<copy_a.len; ++i) { /* do windowing */
    if(copy_a.a[ali.num][i] >= 0) {
      left = MAX(j,i-l);
      right = MIN(copy_a.len,i+r);
      append_ali(&copy_a,j,left,&ali); /* intermediate segment is unchanged */
      copy_ali_to_seqs(&copy_a,left,right,&tmp_s);
      msa(&tmp_s,&new_a,par.free_shift && left==0,
                        par.free_shift && right==copy_a.len);
      for(k=0; k<new_a.len; ++k) /* fill level row */
        new_a.a[new_a.num][k] = copy_a.a[copy_a.num][i];
      append_ali(&new_a,0,new_a.len,&ali);
      j = MIN(copy_a.len,i+r);
      free_ali(&new_a); /* was allocated in msa() */
      if(par.o) {
        fputc('*',stderr);
        fflush(stderr);
      }
    }
  }
  append_ali(&copy_a,j,copy_a.len,&ali); /* append remaining alignment */
  free_seqs(&tmp_s);
  free_ali(&copy_a);
  if(par.o) {
    fputc('\n',stderr);
    fflush(stderr);
  }
} /* realign() */

/*****************************************************************************
 * GLOBAL FUNCTION DEFINITIONS
 *****************************************************************************/
int dcalign(int stopsize, int windowsize)
{
  extern parameters par;
  extern sequences seqs;
  extern alignment ali;

  int p,q,lenmax,i,j,k, *glist,mem;
  tuple lower,upper;

#ifdef WRITE_STATISTICS
  timeChat = timeCmin = 0;
#endif

  if(par.spantree) {
    /* order sequences according to slist */
    ALLOC(glist,int,seqs.num); /* create list of goal positions */
    for(p=0; p<seqs.num; ++p)
      glist[seqs.slist[p]] = p;

    for(p=seqs.num-1; p>0; --p) /* bubble sort of g (and simultaneously seqs)*/
      for(q=0; q<p; ++q)
        if(glist[q]>glist[q+1]) {
          mem = glist[q];
          glist[q] = glist[q+1];
          glist[q+1] = mem;
          SWAP_SEQS(seqs,q,q+1);
/*           SWAP_ALI(ali,q,q+1); not necessary, since empty, yet */
        }
  }

  /* allocate memory */
  ALLOC(lower,int,seqs.num);
  ALLOC(upper,int,seqs.num);
  for(lenmax=INT_MIN,p=0; p<seqs.num; ++p) {
    lower[p] = 0;
    upper[p] = seqs.len[p];
    lenmax = MAX(lenmax,seqs.len[p]);
  }

  /* initializations for calc_cut() */
  ALLOC(lo,int,seqs.num); /* are used only from 1 to seqs.num-1 */
  ALLOC(uo,int,seqs.num);
  ALLOC(olo,int,seqs.num);
  ALLOC(olen,int,seqs.num);
  ALLOC(tmp_cut,int,seqs.num);
  ALLOC(lower_tmp_cut,int,seqs.num);
  ALLOC(upper_tmp_cut,int,seqs.num);
  ALLOC(c,double,seqs.num);
  ALLOC(down,BOOL,seqs.num);

  ALLOC(score,int*,seqs.num);
  ALLOC(col,column*,seqs.num);
  ALLOC(m,matrix*,seqs.num);
  for(p=1; p<seqs.num; ++p) {
    ALLOC(score[p],int,seqs.num);
    ALLOC(col[p],column,seqs.num);
    ALLOC(m[p],matrix,seqs.num);
    for(q=0; q<p; ++q)
      ALLOC(col[p][q],int,lenmax+1); /* lenmax since sequences are exchanged */
  }
  /* matrices are allocated in calc_Chat() */

#ifdef PRIMITIVE_CYCLE
  REALLOC(lo,int,seqs.num+1); /* lo is misused as cycle-array */
  calc_cycle(&seqs,lo);
  lo[seqs.num] = lo[0]; /* copy 0'th entry to the end */
  for(p=0; p<seqs.num; ++p)
    for(q=0; q<seqs.num; ++q)
      seqs.wf[p][q] = 0.0;
  for(p=0; p<seqs.num; ++p)
    seqs.wf[lo[p]][lo[p+1]] = seqs.wf[lo[p+1]][lo[p]] = 1.0;
  if(par.o) {
    write_weights(stderr,&seqs);
    fflush(stderr);
  }
#endif

  /* print dots */
  if(par.o) {
    fprintf(stderr,"Computing DC Alignment.\n");
    for(i=lenmax,j=1; i>stopsize; i/=2,j*=2) /* print $2^\ceil{n/L}-1$ dots */
      for(k=0; k<j; ++k)
        fprintf(stderr,"..");
    fputc('\n',stderr);
    fflush(stderr);
  }

  ali.len = 0; /* reset alignment */
  ali.msa_approx = FALSE;

  dca(lower,upper,0); /* call recursive function */
  if(par.o) {
    fputc('\n',stderr);
    fflush(stderr);
  }

  if(windowsize > 0)
    realign(windowsize); /* do windowing */

  /* free memory */
  FREE(lower);
  FREE(upper);

  FREE(lo);
  FREE(uo);
  FREE(olo);
  FREE(olen);
  FREE(tmp_cut);
  FREE(lower_tmp_cut);
  FREE(upper_tmp_cut);
  FREE(c);
  FREE(down);

  for(p=1; p<seqs.num; ++p) {
    for(q=0; q<p; ++q)
      FREE(col[p][q]);
    FREE(score[p]);
    FREE(col[p]);
    FREE(m[p]);
  }
  FREE(score);
  FREE(col);
  FREE(m);

#ifdef WRITE_STATISTICS
  fprintf(stderr," %6.2f timeChat %6.2f timeCmin",
          (float)timeChat/CLK_TCK,(float)timeCmin/CLK_TCK);
  fflush(stderr);
#endif

  if(par.spantree) {
    /* order sequences and alignment back */
    for(p=0; p<seqs.num; ++p)
      glist[p] = seqs.slist[p];
  
    for(p=seqs.num-1; p>0; --p) /* bubble sort of g (and simultaneously seqs) */
      for(q=0; q<p; ++q)
        if(glist[q]>glist[q+1]) {
          mem = glist[q];
          glist[q] = glist[q+1];
          glist[q+1] = mem;
          SWAP_SEQS(seqs,q,q+1);
          SWAP_ALI(ali,q,q+1);
        }
    FREE(glist);
  }

  return(calc_ali_score(&ali));
} /* dcalign() */

/****** EOF (dcali.c) ********************************************************/
