/*****************************************************************************
 *  FILE   : main.c
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: main.c,v 1.6 1997/07/31 16:19:13 jones Exp jones $
 *****************************************************************************/

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* FILE,fprintf(),sprintf(),sscanf() */
#include <string.h> /* strncpy(),strchr(),strlen() */
#include <stdlib.h> /* atof(),exit() */
#include <sys/types.h> /* clock_t */
#include <sys/times.h> /* struct tms,times() */

#include "global.h" /* data types,defaults,macros */
#include "seqs.h" /* function prototypes */
#include "scores.h" /* function prototypes */
#include "dcali.h" /* dcalign() */

/*****************************************************************************
 * GLOBAL VARIABLES
 *****************************************************************************/
parameters par;
sequences seqs;
alignment ali;
scores scrs;

/*****************************************************************************
 * MACROS
 *****************************************************************************/
#define USAGE_ERROR {\
                      fprintf(stderr,"Usage: %s %s\n",argv[0],USAGE);\
                      exit(1);\
                    }

/*****************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************/

/*===========================================================================*/
/* -- SET_FILENAME_BASE --
 * Extract name part of p->seqs_filename and set p->seqs_filename_base and
 * alignment filename defaults accordingly */
/*---------------------------------------------------------------------------*/
void set_filename_base(parameters *p)
{
  int i;

  strncpy(p->seqs_filename_base,p->seqs_filename,MAX_LENGTH);
  if(strchr(p->seqs_filename_base,'.')!=NULL) { /* find last dot */
    for(i=strlen(p->seqs_filename_base);
        p->seqs_filename_base[i]!='.' && p->seqs_filename_base[i]!='/';
        --i)
      /* nothing */ ;
    if(p->seqs_filename_base[i] == '.') /* otherwise: '/' */
      p->seqs_filename_base[i] = '\0';
  }
  sprintf(p->aln_filename,"%s.%s",p->seqs_filename_base,ALN_SUFFIX);
  sprintf(p->pir_filename,"%s.%s",p->seqs_filename_base,PIR_SUFFIX);
  sprintf(p->nex_filename,"%s.%s",p->seqs_filename_base,NEX_SUFFIX);
  sprintf(p->dca_filename,"%s.%s",p->seqs_filename_base,DCA_SUFFIX);
} /* set_filename_base() */

/*===========================================================================*/
/* -- PARSE_COMMAND_LINE --
 * Parse command line and save option setting in global parameters 'par' */
/*---------------------------------------------------------------------------*/
void parse_command_line(int argc, char **argv, int *order_start, int *order_end)
{
  int i;
  FILE *fptr;

  par.q = DEFAULT_Q; /* FALSE */
  par.free_shift = DEFAULT_FREE_SHIFT; /* TRUE */
  par.a = DEFAULT_A; /* FALSE */
  par.o = DEFAULT_O; /* TRUE */
  par.weight_intensity = DEFAULT_WEIGHT_INTENSITY; /* 0.0 */
  par.weights_filename[0] = '\0'; /* no default */
  par.spantree = DEFAULT_T; /* FALSE */
  par.scrs_filename[0] = '\0'; /* cost/blosum62 OR cost/dnarna */
  par.l = DEFAULT_L; /* 30 */
  par.w = DEFAULT_W; /* 0 */
  par.format = DEFAULT_FORMAT; /* ALN (= 1) */
 
  for(i=1; i<argc && argv[i][0] == '-'; ++i) {
    switch(argv[i][1]) {
      case 'q': par.q = TRUE;
                break;
      case 'g': par.free_shift = FALSE;
                break;
      case 'a': par.a = TRUE;
                break;
      case 'o': par.o = FALSE;
                break;
      case 'b': if(++i==argc)
                  USAGE_ERROR;
                if((fptr=fopen(argv[i],"r")) != NULL) {
                  FCLOSE(fptr);
                  strncpy(par.weights_filename,argv[i],MAX_LENGTH);
                  par.weight_intensity = -1.0;
                }
                else {
                  par.weight_intensity = atof(argv[i]);
                  if(par.weight_intensity<0.0 || par.weight_intensity>1.0) {
                    fprintf(stderr,"Please provide a weight intensity \
                                    between 0.0 and 1.0.\n");
                    exit(1);
                  }
                }
                break;
      case 't': par.spantree = TRUE;
                *order_start = i+1;
                while(atoi(argv[i+1]) != 0)
                  ++i;
                *order_end = i+1;
                break;
      case 'c': if(++i==argc)
                  USAGE_ERROR;
                strncpy(par.scrs_filename,argv[i],MAX_LENGTH);
                break;
      case 'l': if(++i==argc)
                  USAGE_ERROR;
                par.l = atoi(argv[i]);
                if(par.l <= 0) {
                  fprintf(stderr,
                          "Please provide a stop size larger than 0.\n");
                  exit(1);
                }
                break;
      case 'w': if(++i==argc)
                  USAGE_ERROR;
                par.w = atoi(argv[i]);
                if(par.w < 0) {
                  fprintf(stderr,
                         "Please provide a non-negative window size.\n");
                  exit(1);
                }
                break;
      case 'f': if(++i==argc)
                  USAGE_ERROR;
                par.format = atoi(argv[i]);
                if(par.format!=ALN && par.format!=PIR && par.format!= NEX &&
                   par.format!=DCA) {
                  fprintf(stderr,
                         "Unknown output format %s.\n",argv[i]);
                  exit(1);
                }
                break;
      default:  fprintf(stderr,"Unknown option -%c.\n",argv[i][1]);
                USAGE_ERROR;
    }
  }

  if(i != argc-1)
    USAGE_ERROR;

  strncpy(par.seqs_filename,argv[i],MAX_LENGTH);
  set_filename_base(&par);

} /* parse_command_line() */

/*===========================================================================*/
/* -- MENU --
 * Show menu with current settings and read choice */
/*---------------------------------------------------------------------------*/
int menu(BOOL predefined_sequence_order)
{
  int p, choice;
  char tmp[MAX_LENGTH];

  fprintf(stdout,"\n DCA  Main Menu\n");
  fprintf(stdout,"=================\n");
  fprintf(stdout,"\n 11 sequence file:\t\t%s\n",par.seqs_filename);
  fprintf(stdout,"    (%d sequences: ",seqs.num);
  for(p=0; p<seqs.num; ++p)
    fprintf(stdout,"%d, ",seqs.len[p]);
  fprintf(stdout,"\b\b)\n");
  fprintf(stdout," 12 show sequences\n");
  fprintf(stdout," 13 show weightfactors\n");
  fprintf(stdout," 14 show sequence order\n");

  fprintf(stdout,"\n 21 score on the alphabet:\t%s\n",par.scrs_filename);
  fprintf(stdout," 22 show cost matrix\n");

  fprintf(stdout,"\n 3  use approx. cut positions:\t%s\n",par.a ? "YES"
                                                                : "NO");
  fprintf(stdout,"\n 41 stopsize L:\t\t\t%d\n",par.l);
  fprintf(stdout," 42 window size W:\t\t%d\n",par.w);
  fprintf(stdout," 43 free shift:\t\t\t%s\n",par.free_shift ? "YES"
                                                            : "NO");
  if(par.weight_intensity < 0.0)
    fprintf(stdout," 44 sequence weight file:\t%s\n",par.weights_filename);
  else
    if(par.weight_intensity == 0.0)
      fprintf(stdout," 44 score weighting:\t\tOFF\n");
    else
      fprintf(stdout," 44 weight intensity:\t\t%.2f\n",par.weight_intensity);
  fprintf(stdout," 45 sequence order:\t\t%s\n",
          par.spantree ? (predefined_sequence_order ? "pre-defined" :
                          "by closest neighbor") : "longest sequence first");

  fprintf(stdout,"\n 5 compute D&C Alignment\n");

  if(ali.len > 0) {
    fprintf(stdout,"\n Show alignment\n");
    fprintf(stdout,"     61 in standard format\n");
    fprintf(stdout,"     62 in NBRF/PIR (GDE) format\n");
    fprintf(stdout,"     63 in NEXUS format\n");
    fprintf(stdout,"     64 with cut positions\n");

    fprintf(stdout,"\n Write alignment to file\n");
    fprintf(stdout,"     71 in standard format\n");
    fprintf(stdout,"     72 in NBRF/PIR (GDE) format\n");
    fprintf(stdout,"     73 in NEXUS format\n");
    fprintf(stdout,"     74 with cut positions\n");
  }
  fprintf(stdout,"\n 8 System call\n");
  fprintf(stdout,"\n 9 EXIT\n");
  fprintf(stdout,"\n    enter number please: ");
  FGETS(tmp,MAX_LENGTH,stdin);
  sscanf(tmp,"%d",&choice);
  fputc('\n',stdout);
  return(choice);
} /* menu() */

/*****************************************************************************
 * GLOBAL FUNCTION MAIN
 *****************************************************************************/
int main(int argc, char *argv[])
{
  int i,j,order_start,order_end,item,score,new_l,new_w;
  FILE *fptr;
  char tmp[MAX_LENGTH],command[MAX_LENGTH]="";
  float new_weight_intensity;
  struct tms start,stop;
  clock_t elapsed_time;
  char *t1,t2[MAX_LENGTH]; /* for FOPEN_READABLE_BASE() */

  order_start = order_end = 0;
  parse_command_line(argc,argv,&order_start,&order_end);

  FOPEN_READABLE_BASE(fptr,par.seqs_filename,SEQS_DIR);
  if(fptr == NULL)
    exit(1);
  read_seqs_pir(fptr,&seqs);
  FCLOSE(fptr);
  if(seqs.num == 0)
    exit(1);

  /* prepare alignment */
  alloc_ali(&seqs,&ali);

  if(par.scrs_filename[0] == '\0') {
    sprintf(par.scrs_filename,"%s",
            seqs.type==PROTEIN_TYPE ? DEFAULT_SCRS_PROTEIN_FILENAME :
            (seqs.type==IUPAC_TYPE ? DEFAULT_SCRS_IUPAC_FILENAME :
             DEFAULT_SCRS_UNKNOWN_FILENAME));
  }

  FOPEN_READABLE_BASE(fptr,par.scrs_filename,SCORES_DIR);
  if(fptr == NULL)
      exit(1);
  read_scores(fptr,&scrs);
  FCLOSE(fptr);
  reset_gapcost(&scrs,par.free_shift);
  reset_scores(&seqs,par.weight_intensity,par.weights_filename,par.spantree);

  if(order_start != order_end) {
    for(i=order_start,j=0; i<order_end; ++i,++j)
      seqs.slist[j] = atoi(argv[i])-1;
    if(j != seqs.num)
      USAGE_ERROR;
  }


  if(par.q) { /* quiet mode */
    if(par.o)
      write_weights(stdout,&seqs);
#ifdef WRITE_STATISTICS
    fprintf(stderr,"\"%2d k %5d n",seqs.num,seqs.len[0]);
#endif

    times(&start);
    score = dcalign(par.l,par.w);
    times(&stop);
    elapsed_time = stop.tms_utime-start.tms_utime
                   +stop.tms_cutime-start.tms_cutime;

#ifdef WRITE_STATISTICS
    fprintf(stderr," %8d score %5.2f time",score,(float)elapsed_time/CLK_TCK);
#else
    write_ali(stdout,&ali,par.format);
    fprintf(stdout,"\n");
    if(par.format==DCA || par.o) {
      fprintf(stdout,"Sequences: %s\n",par.seqs_filename);
      fprintf(stdout,"Costfile: %s",par.scrs_filename);
      if(par.free_shift)
        fprintf(stdout,", free shift");
      fprintf(stdout,"\nL = %d, W = %d,",par.l,par.w);
      if(par.weight_intensity == 0.0)
        fprintf(stdout," weights off");
      else
        fprintf(stdout," weight intensity = %.2f",par.weight_intensity);
      if(par.a)
        fprintf(stdout,", approximate cut positions\n");
      else
        fprintf(stdout,"\n");
      fprintf(stdout,"Alignment cost: %d%s\n",score,ali.msa_approx?"*":" ");
      fprintf(stdout,"Elapsed CPU time: %.2f sec.\n\n",
              (float)elapsed_time/CLK_TCK);
    }
#endif
  }

  else { /* interactive mode */

    item = 0;
    while (item != 9) {
      item = menu(order_start!=order_end);
      switch (item) {
        case 11: fprintf(stdout,
                         "\nEnter the name of the sequence file [%s]: ",
                         par.seqs_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp == '\0')
                   strncpy(tmp,par.seqs_filename,MAX_LENGTH);
                 FOPEN_READABLE_BASE(fptr,tmp,SEQS_DIR);
                 if(fptr != NULL) {
                   free_seqs(&seqs);
                   free_ali(&ali);
                   read_seqs_pir(fptr,&seqs);
                   FCLOSE(fptr);
                   alloc_ali(&seqs,&ali);
                   reset_scores(&seqs,par.weight_intensity,
                                par.weights_filename,par.spantree);
                   strncpy(par.seqs_filename,tmp,MAX_LENGTH);
                   set_filename_base(&par);
                 }
                 break;
        case 12: write_seqs(stdout,&seqs);
                 fprintf(stdout,"Press <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 13: write_weights(stdout,&seqs);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 14: write_sequence_order(stdout,&seqs,par.spantree);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 21: fprintf(stdout,
                         "\nEnter the name of the cost matrix file [%s]: ",
                         par.scrs_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp == '\0')
                   strncpy(tmp,par.scrs_filename,MAX_LENGTH);
                 FOPEN_READABLE_BASE(fptr,tmp,SCORES_DIR);
                 if(fptr != NULL) {
                   read_scores(fptr,&scrs);
                   FCLOSE(fptr);
                   reset_gapcost(&scrs,par.free_shift);
                   reset_scores(&seqs,par.weight_intensity,
                                par.weights_filename,par.spantree);
                   ali.len = 0; /* reset alignment */
                   strncpy(par.scrs_filename,tmp,MAX_LENGTH);
                 }
                 break;
        case 22: write_scores(stdout,&scrs);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case  3: par.a = !par.a;
                 ali.len = 0; /* reset alignment */
                 break;
        case 41: do {
                   new_l = par.l;
                   fprintf(stdout,"\nEnter new stop size L [%d]: ",new_l);
                   FGETS(tmp,MAX_LENGTH,stdin);
                   if(*tmp != '\0') {
                     new_l = atoi(tmp);
                     if(new_l <= 0)
                       fprintf(stdout,
                               "Please provide a positive integer value.\n");
                     else {
                       par.l = new_l;
                       ali.len = 0; /* reset alignment */
                     }
                   }
                 } while(new_l <= 0);
                 break;
        case 42: do {
                   new_w = par.w;
                   fprintf(stdout,"\nEnter new window size W [%d]: ",new_w);
                   FGETS(tmp,MAX_LENGTH,stdin);
                   if(*tmp != '\0') {
                     new_w = atoi(tmp);
                     if(new_w < 0)
                       fprintf(stdout,
                             "Please provide a non-negative integer value.\n");
                     else {
                       par.w = new_w;
                       ali.len = 0; /* reset alignment */
                     }
                   }
                 } while(new_w < 0);
                 break;
        case 43: par.free_shift = !par.free_shift;
                 reset_gapcost(&scrs,par.free_shift);
                 reset_scores(&seqs,par.weight_intensity,
                              par.weights_filename,par.spantree);
                 ali.len = 0; /* reset alignment */
                 break;
        case 44: do {
                   new_weight_intensity = par.weight_intensity;
                   if(par.weight_intensity < 0.0)
                     fprintf(stdout,
                     "\nEnter new weights filename or weight intensity [%s]: ",
                             par.weights_filename);
                   else
                     fprintf(stdout,
 "\nEnter new weight intensity (off=0.0,max=1.0) or weights filename [%.2f]: ",
                             new_weight_intensity);
                   FGETS(tmp,MAX_LENGTH,stdin);
                   if(*tmp != '\0') {
                     if((fptr=fopen(tmp,"r")) != NULL) {
                       FCLOSE(fptr);
                       strncpy(par.weights_filename,tmp,MAX_LENGTH);
                       par.weight_intensity = -1.0;
                       reset_scores(&seqs,par.weight_intensity,
                                    par.weights_filename,par.spantree);
                       ali.len = 0; /* reset alignment */
                       new_weight_intensity = 0.0; /* exit do loop */
                     }
                     else {
                       new_weight_intensity = atof(tmp);
                       if(new_weight_intensity<0.0||new_weight_intensity>1.0)
                         fprintf(stdout,
                              "Please provide a value between 0.0 and 1.0.\n");
                       else {
                         par.weight_intensity = new_weight_intensity;
                         reset_scores(&seqs,par.weight_intensity,
                                      par.weights_filename,par.spantree);
                         ali.len = 0; /* reset alignment */
                       }
                     }
                   }
                 } while(*tmp!='\0' && (new_weight_intensity<0.0 ||
                                        new_weight_intensity>1.0));
                 break;
        case 45: par.spantree = !par.spantree;
                 order_start = order_end = 0; /* reset these for menu */
                 reset_scores(&seqs,par.weight_intensity,
                              par.weights_filename,par.spantree);
                 ali.len = 0; /* reset alignment */
                 break;
        case  5: if(seqs.num > 1) {
                   times(&start);
                   score = dcalign(par.l,par.w);
                   times(&stop);
                   elapsed_time = stop.tms_utime-start.tms_utime+
                                  stop.tms_cutime-start.tms_cutime;
                   write_ali(stdout,&ali,par.format);
                   fprintf(stdout,"Alignment cost: %d\n",score);
                   fprintf(stdout,"Elapsed CPU time: %.2f sec.\n",
                           (float)elapsed_time/CLK_TCK);
                   fprintf(stdout,"\nPress <ENTER> to continue.\n");
                   FGETS(tmp,MAX_LENGTH,stdin);
                 }
                 else
                   fprintf(stderr,"You can't align a single sequence!\n");
                 break;
        case 61: write_ali(stdout,&ali,1);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 62: write_ali(stdout,&ali,2);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 63: write_ali(stdout,&ali,3);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 64: write_ali(stdout,&ali,4);
                 fprintf(stdout,"\n");
                 fprintf(stdout,"Sequences: %s\n",par.seqs_filename);
                 fprintf(stdout,"Costfile: %s",par.scrs_filename);
                 if(par.free_shift)
                   fprintf(stdout,", free shift");
                 fprintf(stdout,"\nL = %d, W = %d,",par.l,par.w);
                 if(par.weight_intensity == 0.0)
                   fprintf(stdout," weights off");
                 else
                   fprintf(stdout," weight intensity = %.2f",
                           par.weight_intensity);
                 if(par.a)
                   fprintf(stdout,", approximate cut positions\n");
                 else
                   fprintf(stdout,"\n");
                 fprintf(stdout,"Alignment cost: %d%s\n",score,
                         ali.msa_approx?"*":" ");
                 fprintf(stdout,"Elapsed CPU time: %.2f sec.\n\n",
                         (float)elapsed_time/CLK_TCK);
                 fprintf(stdout,"\nPress <ENTER> to continue.\n");
                 FGETS(tmp,MAX_LENGTH,stdin);
                 break;
        case 71: fprintf(stdout,"\nEnter a name for the output file [%s]: ",
                         par.aln_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp != '\0')
                   strncpy(par.aln_filename,tmp,MAX_LENGTH);
                 FOPEN_WRITEABLE(fptr,par.aln_filename);
                 if(fptr != NULL) {
                   write_ali(fptr,&ali,1);
                   FCLOSE(fptr);
                   fprintf(stdout,"Alignment written to file \"%s\"\n",
                           par.aln_filename);
                 }
                 break;
        case 72: fprintf(stdout,"\nEnter a name for the output file [%s]: ",
                         par.pir_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp != '\0')
                   strncpy(par.pir_filename,tmp,MAX_LENGTH);
                 FOPEN_WRITEABLE(fptr,par.pir_filename);
                 if(fptr != NULL) {
                   write_ali(fptr,&ali,2);
                   FCLOSE(fptr);
                   fprintf(stdout,"Alignment written to file \"%s\"\n",
                           par.pir_filename);
                 }
                 break;
        case 73: fprintf(stdout,"\nEnter a name for the output file [%s]: ",
                         par.nex_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp != '\0')
                   strncpy(par.nex_filename,tmp,MAX_LENGTH);
                 FOPEN_WRITEABLE(fptr,par.nex_filename);
                 if(fptr != NULL) {
                   write_ali(fptr,&ali,3);
                   FCLOSE(fptr);
                   fprintf(stdout,"Alignment written to file \"%s\"\n",
                           par.nex_filename);
                 }
                 break;
        case 74: fprintf(stdout,"Enter a name for the output file [%s]: ",
                         par.dca_filename);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp != '\0')
                   strncpy(par.dca_filename,tmp,MAX_LENGTH);
                 FOPEN_WRITEABLE(fptr,par.dca_filename);
                 if(fptr != NULL) {
                   write_ali(fptr,&ali,4);
                   fprintf(fptr,"Sequences: %s\n",par.seqs_filename);
                   fprintf(fptr,"Costfile: %s",par.scrs_filename);
                   if(par.free_shift)    
                     fprintf(fptr,", free shift");    
                   fprintf(fptr,"\nL = %d, W = %d,",par.l,par.w);    
                   if(par.weight_intensity == 0.0)    
                     fprintf(fptr," weights off\n");    
                   else    
                     fprintf(fptr," weight intensity = %.2f\n",
                             par.weight_intensity);
                   fprintf(fptr,"Alignment cost: %d%s\n",score,
                           ali.msa_approx?"*":" ");
                   fprintf(fptr,"Elapsed CPU time: %.2f sec.\n\n",
                           (float)elapsed_time/CLK_TCK);    
                   FCLOSE(fptr);
                   fprintf(stdout,"Alignment written to file \"%s\"\n",
                           par.dca_filename);
                 }
                 break;
        case  8: if(*command=='\0')
                   fprintf(stdout,"\nEnter system command: ");
                 else
                   fprintf(stdout,"\nEnter system command [%s]: ",command);
                 FGETS(tmp,MAX_LENGTH,stdin);
                 if(*tmp != '\0')
                   strncpy(command,tmp,MAX_LENGTH);
                 if(*command != '\0') {
                   system(command);
                   fprintf(stdout,"\nPress <ENTER> to continue.\n");
                   FGETS(tmp,MAX_LENGTH,stdin);
                 }
                 break;
        case  9: break;
        default: fprintf(stderr,"Illegal choice \"%d\".\n\n",item);
                 item = 0;
                 break;
      }
    }
  }
  return(0);
} /* main() */
/****** EOF (main.c) *********************************************************/
