/*****************************************************************************
 *  FILE   : seqs.c
 *  AUTHOR : Jens Stoye (und auch ein bischen Soeren Perrey)
 *  DATE   : 05/1997
 *  RCS-ID : $Id: seqs.c,v 1.6 1997/10/09 17:09:26 jones Exp jones $
 *****************************************************************************
 *  Functions regarding sequences
 *****************************************************************************/

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* FILE,fprintf(),fputc(),feof() */
#include <string.h> /* strstr(),strdup() */
#include <ctype.h> /* isprint(),isspace() */
#include <float.h> /* FLT_MAX */
#include <limits.h> /* INT_MAX,INT_MIN */
#include <sys/types.h> /* getpid(),stat() */
#include <unistd.h> /* getpid() */
#include <sys/stat.h> /* stat() */
#include <errno.h> /* return status of stat */

/* ++pac */
#ifdef _PLAN9
// for strdup()
#include <bsd.h>
#endif


#include "global.h" /* data types,macros */
#include "scores.h" /* encode(),decode() */
#include "dypr.h" /* calc_column() */
#include "seqs.h" /* function prototypes */

/*****************************************************************************
 * GLOBAL FUNCTION DEFINITIONS
 *****************************************************************************/
void alloc_seqs(sequences *s1, sequences *s2)
{
  int p;

  s2->num = s1->num;
  s2->type = s1->type;
  ALLOC(s2->name,char*,s2->num);
  ALLOC(s2->len,int,s2->num);
  ALLOC(s2->s,sequence,s2->num);
  ALLOC(s2->pwscore,int*,s2->num);
  ALLOC(s2->wf,float*,s2->num);
  ALLOC(s2->slist,int,s2->num);
  for(p=0; p<s2->num; ++p) {
    s2->name[p] = strdup(s1->name[p]);
    ALLOC(s2->s[p],letter,s1->len[p]);
    ALLOC(s2->pwscore[p],int,s2->num);
    ALLOC(s2->wf[p],float,s2->num);
  }

} /* alloc_seqs() */

/*----------------------------------------------------------------------------*/
void free_seqs(sequences *s)
{
  int p;

  for(p=0; p<s->num; ++p) {
    FREE(s->name[p]);
    FREE(s->s[p]);
    FREE(s->pwscore[p]);
    FREE(s->wf[p]);
  }
  FREE(s->name);
  FREE(s->len);
  FREE(s->s);
  FREE(s->pwscore);
  FREE(s->wf);
  FREE(s->slist);

} /* free_seqs() */

/*---------------------------------------------------------------------------*/
void read_seqs_pir(FILE *fptr, sequences *s)
{
  int p;
  char seq_name[MAX_LENGTH],*d,c;
  letter l;

  s->num = 0;
  ALLOC(s->name,char*,1);
  ALLOC(s->len,int,1);
  ALLOC(s->s,sequence,1);

  s->len[0] = 0;
  ALLOC(s->s[0],letter,1);
  
  if(feof(fptr)) {
    fprintf(stderr,"Empty file.\n");
    return;
  }

  /* read first non-space line */
  while(!feof(fptr) && (l=encode(c=fgetc(fptr)))==CODE_ERROR)
    /* nothing */ ;
  if(l != CODE_NEW_SEQUENCE) {
    fprintf(stderr,"Sequence format not accepted by DCA. (Use readseq)\n");
    return;
  }

  s->type = TAG2TYPE(c);

  /* read sequence name */
  FGETS(seq_name,MAX_LENGTH,fptr);
  s->name[0] = strdup(seq_name);

  /* read sequence entries and following sequences */
  while(!(feof(fptr))) {
    l = encode(fgetc(fptr));
    if(l == CODE_NEW_SEQUENCE) { /* begin of next sequence determined */
      if(s->len[s->num] > 0) {
        REALLOC(s->s[s->num],letter,s->len[s->num]); /* reset length */
        ++s->num; /* prepare for next sequence */
        REALLOC(s->name,char*,s->num+1);
        REALLOC(s->len,int,s->num+1);
        REALLOC(s->s,sequence,s->num+1);
        s->len[s->num] = 0;
        ALLOC(s->s[s->num],letter,1);
      }
      FGETS(seq_name,MAX_LENGTH,fptr); /* read next sequence name */
      s->name[s->num] = strdup(seq_name);
    }
    else
      if(l > 0) {
        if(s->len[s->num]%SEQS_ALLOCSTEP == 0) /* allocate more memory */
          REALLOC(s->s[s->num],letter,s->len[s->num]+SEQS_ALLOCSTEP);
        s->s[s->num][s->len[s->num]++] = l; /* append letter */
      }
  } /* while */

  if(s->len[s->num] > 0) {
    REALLOC(s->s[s->num],letter,s->len[s->num]);
    ++s->num;
  }

  /* remove any spaces, unprintable characters, and colons from names */
  for(p=0; p<s->num; ++p)
    for(d=s->name[p]; *d!='\0'; ++d)
      if(!isprint(*d) || isspace(*d) || (*d==':'))
        *d = NAME_SPACER;

  /* allocate memory for pairwise scores and weightfactors */
  ALLOC(s->pwscore,int*,s->num);
  ALLOC(s->wf,float*,s->num);
  ALLOC(s->slist,int,s->num);
  for(p=0; p<s->num; ++p) {
    ALLOC(s->pwscore[p],int,s->num);
    ALLOC(s->wf[p],float,s->num);
  }
} /* read_seqs_pir() */

/*---------------------------------------------------------------------------*/
void write_seqs(FILE *fptr, sequences *s)
{
  int p,x,j;

  for(p=0; p<s->num; ++p) {
    fprintf(fptr,"%c%s\n",TYPE2TAG(s->type),s->name[p]);
    for(x=0; x<s->len[p]; x+=SEQS_OUTPUT_LINE_LEN) {
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<s->len[p]; ++j)
        fputc(decode(s->s[p][j]),fptr);
      fputc('\n',fptr);
    }
    fputc('\n',fptr);
  }
} /* write_seqs() */

/*---------------------------------------------------------------------------*/
void copy_seqs(sequences *s1, tuple l, tuple u, sequences *s2)
{
  int p,q,j,k;

  alloc_seqs(s1,s2);
  for(p=0; p<s2->num; ++p) {
    s2->name[p] = strdup(s1->name[p]);
    s2->len[p] = u[p]-l[p];
    for(j=l[p],k=0; k<s2->len[p]; ++j,++k)
      s2->s[p][k] = s1->s[p][j];
    for(q=0; q<s2->num; ++q)
      s2->wf[p][q] = s1->wf[p][q];
  }
} /* copy_seqs() */

/*----------------------------------------------------------------------------*/
void alloc_ali(sequences *s, alignment *a)
{
  int p,lensum;

  a->num = s->num;
  a->type = s->type;
  ALLOC(a->name,char*,a->num);
  for(lensum=0,p=0; p<a->num; ++p) {
    a->name[p] = strdup(s->name[p]);
    lensum += s->len[p];
  }
  ALLOC(a->a,sequence,a->num+1); /* +1 for level row */
  for(p=0; p<=a->num; ++p) /* with level row */
    ALLOC(a->a[p],letter,lensum);
  a->len = 0;
  a->msa_approx = FALSE;
} /* alloc_ali() */

/*---------------------------------------------------------------------------*/
void free_ali(alignment *a)
{
  int p;

  for(p=0; p<a->num; ++p)
    FREE(a->name[p]);
  FREE(a->name);
  for(p=0; p<=a->num; ++p) /* with level row */
    FREE(a->a[p]);
  FREE(a->a);
} /* free_ali() */

/*---------------------------------------------------------------------------*/
void write_ali_aln(FILE *fptr, alignment *a)
{
  int x,p,j,*count;

  ALLOC(count,int,a->num);
  fprintf(fptr,"\nDC Alignment:\n");
  for(p=0; p<a->num; ++p) {
    fprintf(fptr,"%3d %s\n",p+1,a->name[p]);
    count[p] = 0;
  }
  fputc('\n',fptr);
  for(x=0; x<a->len; x+=SEQS_OUTPUT_LINE_LEN) {
    fprintf(fptr,"     "); /* counter row */
    for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j+10<a->len; j+=10)
      fprintf(fptr,"%10d",j+10);
    fputc('\n',fptr);
    for(p=0; p<a->num; ++p) {
      fprintf(fptr,"%3d  ",p+1);
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<a->len; ++j) {
        fprintf(fptr,"%c",decode(a->a[p][j]));
        if(a->a[p][j]!=GAP)
          ++count[p];
      }
      fprintf(fptr,"%5d\n",count[p]);
    }
    fputc('\n',fptr);
  }
} /* write_ali_aln() */

/*---------------------------------------------------------------------------*/
void write_ali_pir(FILE *fptr, alignment *a)
{
  extern sequences seqs;
  int x,p,j;

  for(p=0; p<a->num; ++p) {
    fprintf(fptr,"%c%-18.15s\n",TYPE2TAG(seqs.type),a->name[p]);
    for(x=0; x<a->len; x+=SEQS_OUTPUT_LINE_LEN) {
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<a->len; ++j)
        fputc(decode(a->a[p][j]),fptr);
      fputc('\n',fptr);
    }
    fputc('\n',fptr);
  }
} /* write_ali_pir() */

/*---------------------------------------------------------------------------*/
void write_ali_nex(FILE *fptr, alignment *a)
{
  extern scores scrs;

  int p,j,x;
 
  fprintf(fptr,"#NEXUS\n");

  fprintf(fptr,"BEGIN taxa;\n");
  fprintf(fptr,"    DIMENSIONS ntax=%d;\n",a->num);
  fprintf(fptr,"    TAXLABELS\n");
  for(p=0; p<a->num; ++p)
    fprintf(fptr,"        %-18.15s\n",a->name[p]);
  fprintf(fptr,"    ;\n");
  fprintf(fptr,"END;\n");

  fprintf(fptr,"BEGIN characters;\n");
  fprintf(fptr,"    DIMENSIONS nchar=%d;\n",a->len);
  fprintf(fptr,"    FORMAT\n");
  fprintf(fptr,"        datatype=%s\n",TYPE2NAME(scrs.type));
  fprintf(fptr,"        missing=%c\n",decode(GAP));
  fprintf(fptr,"        symbols=\"%c ",decode(GAP));
  for(j=0; j<scrs.num; ++j)
    fprintf(fptr,"%c ",decode(scrs.idc[j+1]));
  fprintf(fptr,"\b\"\n");
  fprintf(fptr,"        labels\n");
  fprintf(fptr,"        interleave\n");
  fprintf(fptr,"    ;\n");
  fprintf(fptr,"    MATRIX\n");
  for(x=0; x<a->len; x+=SEQS_OUTPUT_LINE_LEN) {
    for(p=0; p<a->num; ++p) {
      fprintf(fptr,"%-18.15s",a->name[p]);
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<a->len; ++j)
        fprintf(fptr,"%c",decode(a->a[p][j]));
      fputc('\n',fptr);
    }
    fputc('\n',fptr);
  }
  fprintf(fptr,"    ;\n");
  fprintf(fptr,"END;\n");
} /* write_ali_nex() */

/*---------------------------------------------------------------------------*/
void write_ali_dca(FILE *fptr, alignment *a)
{
  int x,p,j;

  for(x=0; x<a->len; x+=SEQS_OUTPUT_LINE_LEN) {
    fprintf(fptr,"%-10.7s","Level");
    for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<a->len; ++j)
      fputc(DECODE_LEVEL(a->a[a->num][j]),fptr);
    fputc('\n',fptr);
    for(p=0; p<a->num; ++p) {
      fprintf(fptr,"%-10.7s",a->name[p]);
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<a->len; ++j)
        fputc(decode(a->a[p][j]),fptr);
      fputc('\n',fptr);
    }
    fputc('\n',fptr);
  }
} /* write_ali_dca() */

/*---------------------------------------------------------------------------*/
void write_ali(FILE *fptr, alignment *a, int f)
{
  switch(f) {
    case ALN: write_ali_aln(fptr,a);
              break;
    case PIR: write_ali_pir(fptr,a);
              break;
    case NEX: write_ali_nex(fptr,a);
              break;
    case DCA: write_ali_dca(fptr,a);
              break;
    default: fprintf(stderr,"Unknown output format %d.\n",f);
  }
} /* write_ali() */

/*---------------------------------------------------------------------------*/
void copy_ali(alignment *a1, alignment *a2)
{
  int p,j;

  a2->num = a1->num;
  a2->type = a1->type;
  a2->len = a1->len;
  ALLOC(a2->name,char*,a2->num);
  for(p=0; p<a2->num; ++p)
    a2->name[p] = strdup(a1->name[p]);
  ALLOC(a2->a,sequence,a2->num+1); /* +1 for Level row */
  for(p=0; p<=a2->num; ++p) {
    ALLOC(a2->a[p],letter,a2->len);
    for(j=0; j<a2->len; ++j)
      a2->a[p][j] = a1->a[p][j];
  }
  a2->msa_approx = a1->msa_approx;
} /* copy_ali() */

/*---------------------------------------------------------------------------*/
void copy_ali_to_seqs(alignment *a, int l, int u, sequences *s)
{
  int p,k;
  letter c;

  s->num = a->num;
  s->type = a->type;
  for(p=0; p<s->num; ++p) {
    s->len[p] = 0;
    for(k=l; k<u; ++k)
      if((c=a->a[p][k]) != GAP)
        s->s[p][s->len[p]++] = c;
  }
} /* copy_ali_to_seqs() */

/*---------------------------------------------------------------------------*/
void append_ali(alignment *a1, int l, int u, alignment *a2)
{
  int k,p;

  for(k=l; k<u; ++k,++a2->len)
    for(p=0; p<=a2->num; ++p)
      a2->a[p][a2->len] = a1->a[p][k];
  a2->msa_approx |= a1->msa_approx;
} /* append_ali() */

/*---------------------------------------------------------------------------*/
void read_ali_msa(FILE *fptr, alignment *a, BOOL optimal)
{
  char l[MAX_LENGTH],*lp;
  int p,j,oldlen;

  do
    lp = fgets(l,MAX_LENGTH,fptr);
  while(lp != NULL && (optimal ? strstr(l,MSA_OPTIMAL_TEXT)
                               : strstr(l,MSA_HEURISTIC_TEXT)) == NULL);

  a->type = UNKNOWN_TYPE;

  if(lp == NULL) {
    a->num = 0; /* return empty alignment */
    a->len = 0;
    return;
  }

  ALLOC(a->name,char*,1);
  ALLOC(a->a,sequence,1);

  FGETS(l,MAX_LENGTH,fptr); /* read empty line */
  if(!optimal)
    FGETS(l,MAX_LENGTH,fptr); /* read extra line */

  FGETS(l,MAX_LENGTH,fptr); /* read first line of alignment */
  a->len = strlen(l);
  for(a->num=0; l!=NULL && l[0]!='\0'; ++a->num) {
    REALLOC(a->name,char*,a->num+1);
    ALLOC(a->name[a->num],char,1); /* dummy, to make it FREEable */
    REALLOC(a->a,sequence,a->num+1);
    ALLOC(a->a[a->num],letter,a->len);
    for(j=0; j<a->len; ++j)
      a->a[a->num][j] = encode(l[j]);
    FGETS(l,MAX_LENGTH,fptr); /* read next line */
  }

  for(fgets(l,MAX_LENGTH,fptr);
      /* in case -g option is set, an additional test "l[0]!='\n'" would be
       * necessary! */
      (optimal ? strstr(l,MSA_END_GAPS_TEXT) : strstr(l,MSA_TIME_TEXT))==NULL;
      fgets(l,MAX_LENGTH,fptr)) {
    if(!optimal)
      fgets(l,MAX_LENGTH,fptr); /* read extra line */
    oldlen = a->len;
    a->len += strlen(l)-1; /* without newline character */
    for(p=0; p<a->num; ++p) {
      REALLOC(a->a[p],letter,a->len);
      for(j=oldlen; j<a->len; ++j)
        a->a[p][j] = encode(l[j-oldlen]);
      fgets(l,MAX_LENGTH,fptr); /* read next line */
    }
  }

  REALLOC(a->a,int*,a->num+1); /* add level row */
  ALLOC(a->a[a->num],int,a->len);
  for(j=0; j<a->len; ++j)
    a->a[a->num][j] = optimal ? MSA_OPTIMAL_LEVEL : MSA_HEURISTIC_LEVEL;
  a->msa_approx = !optimal;
} /* read_ali_msa() */

/*---------------------------------------------------------------------------*/
void msa(sequences *s, alignment *a, BOOL left, BOOL right)
{
  extern scores scrs;

  int p,x,j;
  FILE *fptr;
  char command[MAX_LENGTH],
       msa_input_filename[MAX_LENGTH], msa_force_filename[MAX_LENGTH],
       msa_cost_filename[MAX_LENGTH], msa_output_filename[MAX_LENGTH];

  sprintf(msa_input_filename,"%s.%d",MSA_INPUT_FILENAME,(int)getpid()),
  sprintf(msa_force_filename,"%s.%d",MSA_FORCE_FILENAME,(int)getpid());
  sprintf(msa_cost_filename,"%s.%d",MSA_COST_FILENAME,(int)getpid());
  sprintf(msa_output_filename,"%s.%d",MSA_OUTPUT_FILENAME,(int)getpid());

  /* write sequences with additional letters */
  FOPEN_WRITEABLE(fptr,msa_input_filename);
  for(p=0; p<s->num; ++p) {
    fprintf(fptr,"%c%s\n",UNKNOWN_TAG,s->name[p]);
    if(!left)
      fprintf(fptr,"%c\n",decode(scrs.idc[scrs.num])); /* any letter */
    for(x=0; x<s->len[p]; x+=SEQS_OUTPUT_LINE_LEN) {
      for(j=x; j<x+SEQS_OUTPUT_LINE_LEN && j<s->len[p]; ++j)
        fputc(decode(s->s[p][j]),fptr);
      fputc('\n',fptr);
    }
    if(!right)
      fprintf(fptr,"%c\n",decode(scrs.idc[scrs.num])); /* any letter */
    fprintf(fptr,"\n");
  }
  FCLOSE(fptr);

  /* write force file (empty, if !left && !right) */
  FOPEN_WRITEABLE(fptr,msa_force_filename);
  if(!left) {
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"%d ",p+1);
    fprintf(fptr,"S ");
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"%d ",1);
    fprintf(fptr,"L 1\n");
  }
  if(!right) {
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"%d ",p+1);
    fprintf(fptr,"S ");
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"%d ",s->len[p]+(left ? 1 : 2));
    fprintf(fptr,"L 1\n");
  }
  FCLOSE(fptr);

  /* write cost file (because of msa error) */
  FOPEN_WRITEABLE(fptr,msa_cost_filename);
  write_scores_msa(fptr,&scrs);
  FCLOSE(fptr);

/* PROBLEM: this would be nice but I don't know how it is done.
  if((j=stat("msa",&stat_buf))!=0 || stat_buf.st_mode ...) {
    printf("stat returned %d\n",j);
    perror("stat");
    printf("stat errno = %d\n",errno);
    fprintf(stderr,"\nCan't find executable %s!\n",MSA_CALL);
    exit(1);
  }
*/

  sprintf(command,"/bin/sh -c \"%s -o -b -f %s -c %s %s > %s\" 2>/dev/null",
          MSA_CALL,
          msa_force_filename,msa_cost_filename,
          msa_input_filename,msa_output_filename);
  system(command);
  if((fptr=fopen(msa_output_filename,"r")) == NULL) {
    fprintf(stderr,"\nCan't execute %s.\n",MSA_CALL);
    exit(1);
  }
  read_ali_msa(fptr,a,TRUE); /* read optimal alignment from msa output */
  FCLOSE(fptr);

  if(a->num == 0) { /* msa did not find optimal alignment */
    REMOVE(msa_output_filename);
    sprintf(command,
            "/bin/sh -c \"%s -o -b -m -f %s -c %s %s > %s\" 2>/dev/null",
            MSA_CALL,
            msa_force_filename,msa_cost_filename,
            msa_input_filename,msa_output_filename);
    system(command);
    if((fptr=fopen(msa_output_filename,"r")) == NULL) {
      fprintf(stderr,"\nCan't execute %s.\n",MSA_CALL);
      exit(1);
    }
    read_ali_msa(fptr,a,FALSE); /* read heuristic alignment */
    FCLOSE(fptr);

    if(a->num == 0) { /* could not find heuristic alignment, either */
      fprintf(stderr,"Incorrect msa run with \"%s\".\n",msa_input_filename);
      exit(1);
    }
  }

  /* remove additional letters */
  if(!left) {
    /* for security reasons: delete first letter in each row */
    for(p=0; p<a->num; ++p) {
      for(j=0; a->a[p][j]==GAP; ++j)
        /* nothing */ ;
      a->a[p][j] = GAP;
    }
    for(p=0; p<=a->num; ++p) /* shift alignment (with level row) */
      for(j=1; j<a->len; ++j)
        a->a[p][j-1] = a->a[p][j];
    --a->len;
  }
  if(!right) {
    /* for security reasons: delete last letter in each row */
    for(p=0; p<a->num; ++p) {
      for(j=a->len-1; a->a[p][j]==GAP; --j)
        /* nothing */ ;
      a->a[p][j] = GAP;
    }
    --a->len; /* decrement alignment length */
  }

  REMOVE(msa_input_filename);
  REMOVE(msa_force_filename);
  REMOVE(msa_cost_filename);
  REMOVE(msa_output_filename);
} /* msa() */

/*---------------------------------------------------------------------------*/
int calc_ali_score(alignment *a)
{
  extern sequences seqs;
  extern scores scrs;

  int p,q,k,score,pwscore;

  for(score=0,p=1; p<a->num; ++p)
    for(q=0; q<p; ++q) {
      pwscore = 0;
      for(k=0; k<a->len; ++k) /* calculate score without gap cost */
        pwscore += scrs.scr[a->a[p][k]][a->a[q][k]];
      for(k=0; k<a->len && a->a[p][k]==GAP && a->a[q][k]==GAP; ++k)
        /* nothing */ ;
      if(k<a->len && ((a->a[p][k]==GAP) + (a->a[q][k]==GAP)) == 1) /* xor */
        pwscore += scrs.endgapcost;
      for(++k; k<a->len; ++k)
        if((((a->a[p][k]==GAP) + (a->a[q][k]==GAP)) == 1) &&
           ((((a->a[p][k]==GAP) + (a->a[p][k-1]==GAP)) == 1) ||
            (((a->a[q][k]==GAP) + (a->a[q][k-1]==GAP)) == 1)))
          pwscore += scrs.gapcost;
      for(k=a->len-1; k>=0 && a->a[p][k]==GAP && a->a[q][k]==GAP; --k)
        /* nothing */ ;
      if(k>=0 && ((a->a[p][k]==GAP) + (a->a[q][k]==GAP)) == 1)
        pwscore -= scrs.gapcost-scrs.endgapcost;
      score += pwscore*seqs.wf[p][q];
    }
  return(score);
} /* calc_ali_score() */

/*---------------------------------------------------------------------------*/
void reset_scores(sequences *s, float wi, char *wf, BOOL st)
{
  extern parameters par;

  int i,j, p,q, maxlen,minscore,maxscore, center,
      minsum,losum,*sum;
  float dist,dist1;
  column col;
  BOOL *selected;
  FILE *fptr;

  /* if necessary calculate pairwise alignment scores */
  if(wi>0.0 || st) {
    for(maxlen=0,p=1; p<s->num; ++p)
      maxlen = MAX(maxlen,s->len[p]);

    if(par.o) {
      fprintf(stderr,"\nComputing pairwise distances.\n");
      for(i=0; i<s->num*(s->num-1)/2; ++i)
        fputc('.',stderr);
      fputc('\n',stderr);
    }

    ALLOC(col,int,maxlen+1);
    for(minscore=INT_MAX,maxscore=INT_MIN,p=1; p<s->num; ++p)
      for(q=0; q<p; ++q) {
        if(par.o)
          fputc('*',stderr);
        s->pwscore[p][q] = s->pwscore[q][p]
                         = calc_column(p,q,0,0,s->len[p],s->len[q],0,col)*
                           NORMALIZE_FACTOR/MIN(s->len[p],s->len[q]);
        minscore = MIN(minscore,s->pwscore[p][q]);
        maxscore = MAX(maxscore,s->pwscore[p][q]);
      }
    FREE(col);

    if(par.o)
      fputc('\n',stderr);
  }
  else /* make weights undefined */
    for(p=0; p<s->num; ++p)
      for(q=0; q<s->num; ++q)
        s->pwscore[p][q] = s->pwscore[q][p] = -1; /* UNDEF */


  /* if wi negative, read weights from file */
  if(wi < 0.0) {
    FOPEN_READABLE(fptr,wf);
    read_weights(fptr,s);
    FCLOSE(fptr);
  }
  else /* if wi zero, set weights to 1.0 */
    if(wi == 0.0) {
      for(p=0; p<s->num; ++p)
        for(q=0; q<s->num; ++q)
          s->wf[p][q] = s->wf[q][p] = 1.0;
    }
    else /* if wi positive, compute weights */
      for(p=1; p<s->num; ++p)
        for(q=0; q<p; ++q)
          s->wf[p][q] = s->wf[q][p] = maxscore==0 ? 1.0 : 
                        1.0 - wi*(s->pwscore[p][q]-minscore)/maxscore;

  /* calculate tree-based sequence order */
  if(st) {
    /* allocate memory */
    ALLOC(sum,int,s->num);
    ALLOC(selected,BOOL,s->num);
    for(p=0; p<s->num; ++p) {
      selected[p] = FALSE;
    }
  
    /* find center of gravity */
    for(minsum=INT_MAX,i=0; i<s->num; ++i) {
      for(sum[i]=0,j=0; j<s->num; ++j)
        if(i != j)
          sum[i] += s->pwscore[i][j];
      if(sum[i] < minsum) {
        minsum = sum[i];
        center = i;
      }
    }

    s->slist[0] = center; /* set center sequence to first position */
    selected[center] = TRUE;

    for(p=1; p<s->num-1; ++p) { /* set second to last but one sequence */
      dist = FLT_MAX;
      for(i=0;i<s->num; ++i)
        if(!selected[i]) {
          losum = 0;
          for(j=0; j<s->num; ++j) {
            if(selected[j])
              losum += s->pwscore[i][j];
            else
              if(j != i)
                losum -= s->pwscore[i][j];
          }
          dist1 = ((float)losum/(float)p) -
                  ((float)(sum[i]-losum)/(float)(s->num-p-1));
          dist1 = (float)s->num * dist1 / (float)sum[i];
          if(dist1 < dist) {
            dist = dist1;
            q = i;
          }
        }
      s->slist[p] = q;
      selected[q] = TRUE;
    }

    for(p=0; selected[p]; ++p) /* set remaining sequence to last position */
      /* nothing */ ;
    s->slist[s->num-1] = p;

    /* free memory */
    FREE(sum);
    FREE(selected);

  } /* if st */
  else
    for(p=0; p<s->num; ++p)
      s->slist[p] = p;

} /* reset_scores() */

/*---------------------------------------------------------------------------*/
void read_weights(FILE *fptr, sequences *s)
{
  int p,q;

  for(p=0; p<s->num; ++p)
    for(q=p+1; q<s->num; ++q) { /* diagonal remains undefined */
      FSCANF(fptr,"%f",&s->wf[p][q]);
      s->wf[q][p] = s->wf[p][q];
    }
} /* write_weights() */

/*---------------------------------------------------------------------------*/
void write_weights(FILE *fptr, sequences *s)
{
  int p,q;

  fprintf(fptr,"\nWeight factors:\n");
  for(q=0; q<s->num-1; ++q) /* unusual loop */
    for(p=q+1; p<s->num; ++p) {
      fprintf(fptr,"normalized score[%d][%d] = %5d,",q+1,p+1,s->pwscore[p][q]);
      fprintf(fptr," wf[%d][%d] = %.3f\n",q+1,p+1,s->wf[p][q]);
    }
  fputc('\n',fptr);
} /* write_weights() */

/*---------------------------------------------------------------------------*/
void write_sequence_order(FILE *fptr, sequences *s, BOOL st)
{
  int p,q, sum,minsum,center, longest;

  if(st) {
    fprintf(fptr,"\nThe pairwise sequence distances:\n    ");
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"  -%2d-",p+1);
    fprintf(fptr,"   -sum-\n");
    for(minsum=INT_MAX,p=0; p<s->num; ++p) {
      fprintf(fptr,"-%2d-",p+1);
      for(sum=0,q=0; q<s->num; ++q) {
        if(p == q)
          fprintf(fptr,"   ---");
        else {
          fprintf(fptr," %5d",s->pwscore[p][q]);
          sum += s->pwscore[p][q];
        }
      }
      if(sum < minsum) {
        minsum = sum;
        center = p;
      }
      fprintf(fptr,"%8d\n",sum);
    }
  
    fprintf(fptr,"\nCenter of gravity: %d\n",center+1);
  }
  
  fprintf(fptr,"\nThe internal sequence order is (%s):\n",
          st ? "by closest neighbor" : "longest sequence first");

  if(st)
    for(p=0; p<s->num; ++p)
      fprintf(fptr," %2d",s->slist[p]+1);
  else {
    for(longest=0,p=1; p<s->num; ++p)
      if(s->len[p] > s->len[longest])
        longest = p;
    fprintf(fptr," %2d",longest+1);
    for(p=0; p<s->num; ++p)
      if(p != longest)
        fprintf(fptr," %2d",p+1);
  }
  fprintf(fptr,"\n");

} /* write_sequence_order() */

/*---------------------------------------------------------------------------*/
void calc_cycle(sequences *s, tuple cycle)
{
  int p,q;
  FILE *fptr;
  char command[MAX_LENGTH],
       splits_input_filename[MAX_LENGTH], splits_output_filename[MAX_LENGTH];

  if(s->num < 4) /* trivial cycle */
    for(p=0; p<s->num; ++p)
      cycle[p] = p;
  else { /* call splits */
    sprintf(splits_input_filename,"%s.%d",SPLITS_INPUT_FILENAME,(int)getpid()),
    sprintf(splits_output_filename,"%s.%d",SPLITS_OUTPUT_FILENAME,
            (int)getpid());

    FOPEN_WRITEABLE(fptr,splits_input_filename);
    fprintf(fptr,"#NEXUS\n");

    fprintf(fptr,"BEGIN taxa;\n");
    fprintf(fptr,"    DIMENSIONS ntax=%d;\n",s->num);
    fprintf(fptr,"    TAXLABELS\n");
    for(p=0; p<s->num; ++p)
      fprintf(fptr,"        %-18.15s\n",s->name[p]);
    fprintf(fptr,"    ;\n");
    fprintf(fptr,"END;\n");

    fprintf(fptr,"BEGIN DISTANCES;\n");
    fprintf(fptr,"    DIMENSIONS ntax=%d;\n",s->num);
    fprintf(fptr,"    FORMAT\n");
    fprintf(fptr,"        triangle=lower\n");
    fprintf(fptr,"        diagonal\n");
    fprintf(fptr,"        labels\n");
    fprintf(fptr,"        missing=?\n");
    fprintf(fptr,"    ;\n");
    fprintf(fptr,"    MATRIX\n");
    for(p=0; p<s->num; ++p) {
      fprintf(fptr,"%-18.15s",s->name[p]);
      for(q=0; q<p; ++q)
        fprintf(fptr,"%6d ",s->pwscore[p][q]);
      fprintf(fptr,"%6d\n",0);
    }
    fprintf(fptr,"    ;\n");
    fprintf(fptr,"END;\n");
    FCLOSE(fptr);

    sprintf(command,"echo \"show ST_SPLITS\" | %s %s | grep CYCLE > %s",
            SPLITS_CALL,splits_input_filename,splits_output_filename);
    system(command);

    if((fptr=fopen(splits_output_filename,"r")) == NULL) {
      fprintf(stderr,"\nCan't execute %s.\n",SPLITS_CALL);
      exit(1);
    }
    fscanf(fptr,"    CYCLE ");
    for(p=0; p<s->num; ++p) {
      FSCANF(fptr,"%d",&cycle[p]);
      --cycle[p]; /* start counting with 0 */
    }
    FCLOSE(fptr);

    REMOVE(splits_input_filename);
    REMOVE(splits_output_filename);
  }
} /* calc_cycle() */

/****** EOF (seqs.c) *********************************************************/
