/*****************************************************************************
 *  FILE   : seqs.h
 *  AUTHOR : Jens Stoye
 *  DATE   : 05/1997
 *  RCS-ID : $Id: seqs.h,v 1.2 1997/07/14 16:39:55 jones Exp jones $
 *****************************************************************************/
#ifndef SEQS_H
#define SEQS_H

/*****************************************************************************
 * INCLUDES
 *****************************************************************************/
#include <stdio.h> /* FILE */

#include "global.h" /* data types,BOOL */

/*****************************************************************************
 * FUNCTION PROTOTYPES
 *****************************************************************************/

/*===========================================================================*/
/* -- ALLOC_SEQS --
 * Allocate memory for sequences 's2' of size 's1' and copy 'num' and 'name's
 * of 's1' to 's2' */
/*---------------------------------------------------------------------------*/
void alloc_seqs(
  sequences *s1, /* pattern sequences */
  sequences *s2 /* destination sequences */
);

/*===========================================================================*/
/* -- FREE_SEQS --
 * Free memory of sequences 's' */
/*---------------------------------------------------------------------------*/
void free_seqs(
  sequences *s /* obsolete sequences */
);

/*===========================================================================*/
/* -- READ_SEQS_PIR --
 * Allocate required memory and read sequences in PIR format from 'fptr'
 * into 's' */
/*---------------------------------------------------------------------------*/
void read_seqs_pir(
  FILE *fptr, /* input file pointer */
  sequences *s /* new sequences */
);

/*===========================================================================*/
/* -- WRITE_SEQS --
 * Write sequences 's' in PIR format to 'fptr' */
/*---------------------------------------------------------------------------*/
void write_seqs(
  FILE *fptr, /* output file pointer */
  sequences *s /* sequences */
);

/*===========================================================================*/
/* -- COPY_SEQS --
 * Allocate memory for 's2' and copy subsequences of 's1', between 'l' and 'u'
 * into 's2', as well as weightfactors */
/*---------------------------------------------------------------------------*/
void copy_seqs(
  sequences *s1, /* pattern sequences */
  tuple l, /* lower bounds */
  tuple u, /* upper bounds */
  sequences *s2 /* destination sequences */
);

/*===========================================================================*/
/* -- ALLOC ALI --
 * Allocate memory in 'a' for any alignment of sequences 's' */
/*---------------------------------------------------------------------------*/
void alloc_ali(
  sequences *s, /* pattern sequences */
  alignment *a /* new alignment */
);

/*===========================================================================*/
/* -- FREE_ALI --
 * Free memory of alignment 'a' */
/*---------------------------------------------------------------------------*/
void free_ali(
  alignment *a /* obsolete alignment */
);

/*===========================================================================*/
/* -- WRITE_ALI --
 * Write alignment 'a' to 'fptr' with position counter */
/*---------------------------------------------------------------------------*/
void write_ali(
  FILE *fptr, /* output file pointer */
  alignment *a, /* alignment */
  int f /* output format */
);

/*===========================================================================*/
/* -- COPY_ALI --
 * Allocate memory for 'a2' and copy alignment 'a1' to 'a2' */
/*---------------------------------------------------------------------------*/
void copy_ali(
  alignment *a1, /* pattern alignment */
  alignment *a2 /* destination alignment */
);

/*===========================================================================*/
/* -- COPY_ALI_TO_SEQS --
 * Copy region between bounds 'l' and 'u' of alignment 'a' into 's' without
 * allocating memory */
/*---------------------------------------------------------------------------*/
void copy_ali_to_seqs(
  alignment *a, /* pattern alignment */
  int l, /* lower bound */
  int u, /* upper bound */
  sequences *s /* sequences */
);

/*===========================================================================*/
/* -- APPEND_ALI --
 * Append region between bounds 'l' and 'u' of alignment 'a1' to the alignment
 * already existing in 'a2' without allocating memory. 's2->msa_approx' is
 * 's2->msa_approx OR s1->msa_approx' */
/*---------------------------------------------------------------------------*/
void append_ali(
  alignment *a1, /* pattern alignment */
  int l, /* lower bound */
  int u, /* upper bound */
  alignment *a2 /* alignment where 'a1' is appended */
);

/*===========================================================================*/
/* -- READ_ALI_MSA --
 * Allocate required memory and read optimal (in case 'optimal'==TRUE) or
 * heuristic ('optimal'==FALSE) alignment from 'fptr' in MSA output format.
 * Indicate optimal alignment by MSA_OPTIMAL_LEVEL, heuristic alignment by
 * MSA_HEURISTIC_LEVEL entries in level row. 'a->msa_approx' is set accord-
 * ingly. Sequence names remain undefined.
 * If 'fptr' does not contain any or does contain corrupted msa output, no
 * memory is allocated and an empty alignment with a->num==0 and a->len==0
 * is returned. */
/*---------------------------------------------------------------------------*/
void read_ali_msa(
  FILE *fptr, /* input file pointer */
  alignment *a, /* alignment */
  BOOL optimal /* read optimal vs. heuristical alignment */
);

/*===========================================================================*/
/* -- MSA --
 * Call msa with sequences 's' and return alignment in 'a'. The required
 * memory is allocated in the subprocedure 'read_ali_msa'. If 'left' ('right')
 * == TRUE, allow free shift at this end of the sequences. If msa could not
 * find an optimal alignment with default settings or if msa produced a
 * runtime error, the heuristic alignment produced by msa is returned.
 * This is specified by the value of 'a->msa_approx'. */
/*---------------------------------------------------------------------------*/
void msa(
  sequences *s, /* sequences to be aligned */
  alignment *a, /* resulting alignment */
  BOOL left, /* free shift at left end */
  BOOL right /* free shift at right end */
);

/*===========================================================================*/
/* -- CALC_ALI_SCORE --
 * Calculate weighted (according to global 'seqs.wf') alignment score of 'a'
 * according to global scores 'scrs' */
/*---------------------------------------------------------------------------*/
int calc_ali_score(
  alignment *a /* alignment */
);

/*===========================================================================*/
/* -- RESET_SCORES --
 * If wi < 0.0, load pairwise sequence weights from file.
 * Otherwise, (if necessary) calculate pairwise distance scores 's->pwscore'
 * and from these the weightfactors 's->wf' according to the formula given in
 *   "Toenges, Perrey, Stoye, Dress: A General Method for Fast Multiple
 *    Sequence Alignment, Gene 172, GC33-GC41, 1996." (WEIGHTS_1)
 * or the formula given in
 *   "J. Stoye: Divide-and-Conquer Multiple Sequence Alignment, Report 97-02,
 *    Technische Fakultaet der Universitaet Bielefeld, Abteilung Informations-
 *    technik, Januar 1997."  (WEIGHTS_2).
 * The procedure uses the global scores 'scrs'.
 * If 'st' == TRUE, then compute the sequence order due to their distance
 * in a minimal spanning tree (3 different methods). */
/*---------------------------------------------------------------------------*/
void reset_scores(
  sequences *s, /* sequences */
  float wi, /* weight intensity */
  char *wf, /* weights filename */
  BOOL st /* flag if spantree shall be used for sequence ordering */
);

/*===========================================================================*/
/* -- READ_WEIGHTS --
 * Read pairwise weightfactors 's->wf' from 'fptr' in format like epsilons
 * of MSA; one number after the other: (1,2) (1,3) (1,4) ... (1,n) 
 *                                           (2,3) (2,4) ... (2,n)
 *                                                 (3,4) ... (3,n)
 *                                                       ...
 *                                                           (n-1,n) */
/*---------------------------------------------------------------------------*/
void read_weights(
  FILE *fptr, /* output file pointer */
  sequences *s /* sequences */
);

/*===========================================================================*/
/* -- WRITE_WEIGHTS --
 * Write pairwise weightfactors 's->wf' to 'fptr' */
/*---------------------------------------------------------------------------*/
void write_weights(
  FILE *fptr, /* output file pointer */
  sequences *s /* sequences */
);

/*===========================================================================*/
/* -- WRITE_SEQUENCE_ORDER --
 * Write current internal sequence order to 'fptr'. */
/*---------------------------------------------------------------------------*/
void write_sequence_order(
  FILE *fptr, /* output file pointer */
  sequences *s, /* sequences */
  BOOL spantree /* flag if spantree order or original order */
);

/*===========================================================================*/
/* -- CALC_CYCLE --
 * Use the splits program to order sequences in a cycle, which is returned as
 * a tuple with the sequence indices 0,...,seqs.num-1 re-ordered according to
 * the order of the sequences around the cycle. As pairwise distances, the
 * entries 'pwscore' in 's' are used which must be computed previously. */
/*---------------------------------------------------------------------------*/
void calc_cycle(
  sequences *s, /* sequences */
  tuple cycle /* indexing order of cycle */
);

#endif /* SEQS_H */
/****** EOF (seqs.h) *********************************************************/
